
import logging
logger = logging.getLogger(__name__)

import pip
try:
	import importlib
except ImportError:
	# pip install importlib
	pip.main(["install", "importlib"])
	import importlib
#end try


def import_or_install(package_name, pip_name):
	# pytz.timzone -> from pytz import timezone -> package_name = pytz, from_package = [timezone]
	imp = None
	for try_i in [1,2,3]:
		try:
			err = None
			imp = import_only(package_name)
			break
		except ImportError as e:
			err = e
			logger.debug("Import failed.", exc_info=True)
			upgrade = try_i >= 2  # import failed twice (one after doing a normal install)
			install_only(pip_name, upgrade)
	else:
		raise err
	return imp

def import_only(package_name, module_list=None):
	if not module_list:
		if "." in package_name:
			package_name, module_list = package_name.rsplit('.', 1)
		else:
			module_list = None
	if module_list:
		logger.debug("Trying import: form \"{module_name}\" import \"{module_list}\".".format(module_name=package_name, module_list=module_list))
	else:
		logger.debug("Trying to import module \"{module_name}\".".format(module_name=package_name))

	try:
		imp = importlib.import_module(package_name)
		if module_list and hasattr(imp, module_list):
			imp = getattr(imp, module_list)
			logger.debug("\"{module_list}\" is an attribute of \"{module_name}\".".format(module_name=package_name, module_list=module_list))
		else:
			imp = importlib.import_module(package_name, package=module_list)
			logger.debug("\"{module_list}\" is an module in \"{module_name}\".".format(module_name=package_name, module_list=module_list))
	except ImportError:
		imp = importlib.import_module(package_name, package=module_list)
	return imp


def install_only(pip_name, upgrade=False):
	logger.warn("Installing package '{pip_name}'.\n"
		   "If that fails, install it manually:\n"
		   "pip install {pip_name}\n"
		   "".format(pip_name=pip_name))
	args = ["install", pip_name, "--verbose"]
	if upgrade and not "--upgrade" in args:
		args.append("--upgrade")
	logger.debug("Trying to install \"{pip_name}\" with pip using the following arguments: {pip_args}".format(pip_name=pip_name, pip_args=args))
	return pip.main(args)


# -*- coding: utf-8 -*-
__author__ = 'luckydonald'

import getpass


try:
	input = raw_input
except NameError:
	input = input
raw_input = input



# modified from http://code.activestate.com/recipes/541096-prompt-the-user-for-confirmation/
def confirm(prompt=None, default=False):
	"""prompts for yes or no response from the user. Returns True for yes and
	False for no.

	'resp' should be set to the default value assumed by the caller when
	user simply types ENTER.

	>>> confirm(prompt='Create Directory?', default=True)
	Create Directory? [y]|n:
	True
	>>> confirm(prompt='Create Directory?', default=False)
	Create Directory? [n]|y:
	False
	>>> confirm(prompt='Create Directory?', default=False)
	Create Directory? [n]|y: y
	True

	"""

	if prompt is None:
		prompt = 'Confirm'

	if default is not None:
		if default == True:
			prompt = '%s [%s]|%s: ' % (prompt, 'y', 'n')
		else:
			prompt = '%s [%s]|%s: ' % (prompt, 'n', 'y')
	else:
		prompt += ' y|n: ' # "%s y|n"
	#end if
	while True:
		ans = input(prompt)
		if not ans and default is not None:
			return default
		if ans not in ['y', 'Y', 'n', 'N']:
			print('Please enter y or n.')
			continue
		if ans == 'y' or ans == 'Y':
			return True
		if ans == 'n' or ans == 'N':
			return False
	#end while
#end def


def answer(prompt=None, default=None):
	if prompt is None:
		prompt = 'Input'
	#end if
	if default is not None:
		prompt = '{prompt} [{default}]: '.format(prompt=prompt, default=default)
	else:
		prompt = '{prompt}: '.format(prompt=prompt)
	ans = None
	while not ans:
		ans = input(prompt)
		if not ans and default is not None:
			return default
		#end if
	#end while, when the user finally entered something
	return ans
#end def


def password(prompt=None, default=None):
	"""
	If default is None, it will retry until the user inputs something.
	else set default="" to allow empty inputs.

	:param prompt: defaults to "Password"
	:param default:
	:return:
	"""
	if prompt is None:
		prompt = 'Password'
	if default is not None:
		prompt = '{prompt} [{default}]: '.format(prompt=prompt, default=default)
	else:
		prompt = '{prompt}: '.format(prompt=prompt)
	passw = None
	while not passw:
		passw = getpass.getpass(prompt=prompt)
		if not passw and default is not None:
			return default


# -*- coding: utf-8 -*-
from __future__ import unicode_literals
__author__ = 'luckydonald'
import sys

if sys.version < '3': # python 2.7
    text_type = unicode
    binary_type = str
    native_type = binary_type
    long_int = long
    def to_native(x):
        return to_binary(x)
else: # python 3
    text_type = str
    binary_type = bytes
    native_type = text_type
    long_int = int
    def to_native(x):
        return to_unicode(x)


def to_binary(x):
    if isinstance(x, text_type):
        return x.encode("utf-8")
    else:
        return x


def to_unicode(x):
    if isinstance(x, binary_type):
        return x.decode("utf-8")
    else:
        return x



# -*- coding: utf-8 -*-
__author__ = 'luckydonald'

import base64
from .dependencies import import_or_install
# import Crypto  # pip install pycrypto
import_or_install("Crypto", "pycrypto")
from Crypto import Random  # pip install pycrypto
from Crypto.Cipher import AES
from Crypto.Hash import MD5
# from usersettings import Settings  # pip install usersettings
Settings = import_or_install("usersettings.Settings", "usersettings")

BS = 16
pad = lambda s: s + (BS - len(s) % BS) * chr(BS - len(s) % BS)
un_pad = lambda s: s[:-ord(s[len(s) - 1:])]


class Store(object):
	def __init__(self, settings_name, key=None):
		if not key:
			settings = Settings(settings_name)  # store settings, password etc.
			settings.add_setting("do-not-change!", str, random())
			settings.load_settings()
			settings.save_settings()
			self.key = settings.get("do-not-change!")
		else:
			self.key = key

	def encrypt(self, raw):
		raw = pad(raw)
		iv = Random.new().read(AES.block_size)
		cipher = AES.new(self.key, AES.MODE_CBC, iv)
		return base64.b64encode(iv + cipher.encrypt(raw))

	def decrypt(self, enc):
		enc = base64.b64decode(enc)
		iv = enc[:16]
		cipher = AES.new(self.key, AES.MODE_CBC, iv)
		return un_pad(cipher.decrypt(enc[16:]))


def random():
	return MD5.new(Random.new().read(4)).hexdigest()



# -*- coding: utf-8 -*-
__author__ = 'luckydonald'

import logging
logger = logging.getLogger(__name__)
from .dependencies import import_or_install

import pip  # installing stuff

import tempfile
import mimetypes # get mime types/suffix for DL.
magic = import_or_install("magic", "python-magic")
import os
import errno #exist_ok workaround

temp_dir_name = "luckydonald-utils"

def gettempdir():
	temp_dir = tempfile.gettempdir()
	temp_dir = os.path.join(temp_dir, temp_dir_name)
	#py3
	# os.makedirs(temp_dir, exist_ok=True) #don't raise errors if existent.
	#py2/3 exist_ok workaround
	try:
		os.makedirs(temp_dir)
	except OSError as e:
		if e.errno != errno.EEXIST:
			raise
	#end exist_ok workaround
	return temp_dir


#end def


def get_file_suffix(file_path=None, file_url=None):
	mime = get_file_mime(file_path=file_path, file_url=file_url)
	return guess_extension(mime)


def guess_extension(mime):
	return mimetypes.guess_extension(type=mime or "")


def get_file_mime(file_path=None, file_url=None):
		if file_url:
			url = file_url
		elif file_path:
			from urllib.request import pathname2url
			url = pathname2url(file_path)
		else:
			raise AttributeError("Neither URL (file_url) nor local path (file_path) given.")
		mime = magic.from_file(file_path).decode("utf-8")
		return mime

def do_a_filename(input_file_name):
	output_file_name = input_file_name
	for replacer in [(":",""), ("?",""), ("*",""), ("”","\""),(">",""), ("<",""), ("|","-"), ("\\"," "), ("/"," ")]:
		output_file_name = output_file_name.replace(replacer[0], replacer[1])
	logger.debug("Filename '{old_filename}' is now '{new_filename}'.".format(old_filename=input_file_name, new_filename=output_file_name))
	return output_file_name


import subprocess
import sys

"""
def open_folder(folder_path)
"""  # for different Platforms
if sys.platform == 'darwin':  # Mac OS
	def open_folder(folder_path):
		subprocess.check_call(['open', '--', folder_path])
	def open_file_folder(folder_path):
		subprocess.check_call(['open', '-R', '--', folder_path])
elif sys.platform == 'linux2':  # linux, hopefully has gnome?
	def open_folder(folder_path):
		subprocess.check_call(['gnome-open', '--', folder_path])  # untested.
	def open_file_folder(file_path):
		open_folder(os.path.dirname(file_path))
elif sys.platform == 'win32':  # windows
	def open_folder(folder_path):
		subprocess.check_call(['explorer', folder_path])
	def open_file_folder(file_path):
		try:
			subprocess.check_call(['explorer', '/select,"%s"' % file_path])
		except:
			open_folder(os.path.dirname(file_path))



# -*- coding: utf-8 -*-
__author__ = 'luckydonald'
from .dependencies import import_or_install
from .files import gettempdir
import logging  # pip install luckydonald-utils
logger = logging.getLogger(__name__)

import hashlib
import mimetypes
import os

DictObject = import_or_install("DictObject.DictObject", "DictObject")  # from DictObject import DictObject
magic = import_or_install("magic", "python-magic")  # import magic
requests = import_or_install("requests", "requests")  # import requests
from requests.packages.urllib3.exceptions import HTTPError

HEADERS = {'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X10.69; rv:4458.42) Gecko/4458 Firefox/69.0 Pon3Downloader'}


def get_json(url, objectify=True, **kwargs):
	kwargs.setdefault("headers", HEADERS)
	kwargs.setdefault("verify", False)
	json = requests.get(url, **kwargs).json()
	if objectify:
		return DictObject.objectify(json)
	return json


def download_file(url, used_cached=True, temp_dir=None, return_mime=False, return_buffer=False, progress_bar=False, **requests_kwargs):
	"""

	:param url:
	:param used_cached:
	:param temp_dir:
	:param return_mime:
	:param return_buffer: True: Don't write a file, just return the buffer. False: Write to file, return the files path
	:type  return_buffer: bool
	:return:
	"""
	requests_kwargs.setdefault("headers", HEADERS)
	requests_kwargs.setdefault("verify", False)
	if not return_buffer:
		if not temp_dir:
			temp_dir = gettempdir()
		file_name = url.split("/")[-1]
	#end if not return_buffer
	try:
		logger.debug("DL: Downloading from '{url}'.".format(url=url))
		image_buffer = requests.get(url, **requests_kwargs).content
	except HTTPError as e:
		logger.exception("DL: Error in URL '" + url + "'.\n" + str(e))
		raise
		#return (None, None) if return_mime else None
	except Exception as e:
		logger.exception("DL: Error in Download '" + url + "'.\n" + str(e))
		raise
		#return (None, None) if return_mime else None
	mime = magic.from_buffer(image_buffer, mime=True).decode("utf-8")
	suffix = (mimetypes.guess_extension(mime) if mime else ".unknown") or ".unknown"
	if return_buffer:
		logger.debug("DL: Requested Buffer, not creating/checking dirs/files.")
		if return_mime:
			return image_buffer, mime
		return image_buffer
	else: # -> not return_buffer:
		file_name = str(hashlib.md5(url.encode()).hexdigest()) + suffix
		file_name = os.path.join(temp_dir, file_name)
		if os.path.isfile(file_name):
			if used_cached:
				logger.debug("DL: File exists, using cached: %s" % file_name)
				if return_mime:
					return file_name, mime
				return file_name
			logger.debug("DL: File exists, redownloading: %s" % file_name)
		else:
			logger.debug("DL: File does not exist, downloading: %s" % file_name)
			if not os.path.exists(os.path.dirname(file_name)):
				logger.debug("DL: Download Folder does not exists. Creating.")
				os.makedirs(os.path.dirname(file_name))
		logger.debug("DL: Writing download from '%s' to file '%s'" % (url, file_name))
		try:
			with open(file_name, 'wb') as f:
				f.write(image_buffer)
		except Exception as e:
			logger.error("DL: Error in writing download to disk: '" + url + "' to '" + file_name + "'.\n" + str(e))
			file_name = None
		if return_mime:
			return file_name, mime
		return file_name
	#end if-else


# -*- coding: utf-8 -*-
from luckydonaldUtils import VERSION

__author__ = 'luckydonald'

from setuptools import setup

long_description = """A collection of utilities I use across different projects"""

# http://peterdowns.com/posts/first-time-with-pypi.html
# $ python setup.py register -r pypi
# $ python setup.py sdist upload -r pypi

install_requires = ["pip"]
try:
    import importlib
except ImportError:
    install_requires.append('importlib')

setup(
	name="luckydonald-utils",
	packages=['luckydonaldUtils'],
	version=VERSION,
	author="luckydonald",
	author_email="code@luckydonald.de",
	description=long_description,
	license="BSD",
	keywords="example documentation tutorial",
	url="https://github.com/luckydonald/python-utils",
	install_requires=install_requires,
	long_description=long_description,
	classifiers=[
		"Development Status :: 4 - Beta",
		"Intended Audience :: Developers",
		"Topic :: Utilities",
		"License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
	],
)



# -*- coding: utf-8 -*-
__author__ = 'luckydonald'

CHARS_UNESCAPED = ["\\", "\n", "\r", "\t", "\b", "\a", "'"]
CHARS_ESCAPED = ["\\\\", "\\n", "\\r", "\\t", "\\b", "\\a", "\\'"]


def text_split(text, limit, max_parts=None):
	"""
	Split a text usefull.
	Prefer splitting at linebreaks, sentences, words.

	:type text: str
	:type limit: int
	:param max_parts: If not None: In how many parts it should be splitted. The last element might be longer as the given limit.
	:type  max_parts: None | int
	:return:
	"""
	if not isinstance(text, str):
		raise TypeError("text is not str but {}".format(type(text)))
	assert isinstance(limit, int)
	assert max_parts is None or isinstance(max_parts, int)
	text = text.strip()
	if len(text) == 0 or limit == 0:
		return []
	if max_parts == 1:
		return [text]
	assert max_parts is None or max_parts > 1
	#LinkRanges lnkRanges = textParseLinks(text, TextParseLinks | TextParseMentions | TextParseHashtags);
	#currentLink = 0
	#lnkCount = lnkRanges.size()
	half = limit // 2
	good_level = 0
	good = 0
	end = len(text)
	unicode_i = 0
	for i in range(0, end):
		char = text[i]
		#while (currentLink < lnkCount && ch >= lnkRanges[currentLink].from + lnkRanges[currentLink].len) {
		#	++currentLink;
		#}

		#bool inLink = (currentLink < lnkCount) && (ch > lnkRanges[currentLink].from) && (ch < lnkRanges[currentLink].from + lnkRanges[currentLink].len);
		if unicode_i > half:
			if char=="\n" :
				if i + 1 < end and text[i+1] == "\n" and good_level <= 7:
					good_level = 7
					good = i
				elif good_level <= 6:
					good_level = 6
					good = i
			elif char.isspace():
				if is_sentence_end(char) and good_level <= 5:
					good_level = 5
					good = i
				elif is_sentence_part_end(text[i-1]) and good_level <= 4:
					good_level = 4
					good = i
				elif good_level <= 3:
					good_level = 3
					good = i
			elif is_word_separator(text[i-1]) and good_level <= 2:
				good_level = 2
				good = i
			elif good_level <= 1:
				good_level = 1
				good = i
			#end if
		#end if
		if unicode_i >= limit:
			if max_parts == 1:
				return [text.strip()]
			first_text = text[:good]
			text = text[good:]
			parts = [first_text.strip(),]
			if max_parts is None:
				parts.extend(text_split(text, limit))
			elif max_parts > 1:
				parts.extend(text_split(text, limit, max_parts=max_parts-1))
			else:
				raise ArithmeticError("max_split reached illegal state: {}".format(max_parts))
			return parts
		#end if
		if char in CHARS_UNESCAPED:
			unicode_i += 2
		else:
			unicode_i += len(char.encode('utf-8'))
	#end for
	return [text]
#end def
"""
		if len(text)> MAX_TEXT_LENGTH:
			last_space = text.rfind(" ", MAX_TEXT_LENGTH-40, MAX_TEXT_LENGTH)
			last_newline = text.rfind("\n", MAX_TEXT_LENGTH-40, MAX_TEXT_LENGTH)
			if abs(last_newline-last_space) < 20:
				split_at = last_newline
			else:
				split_at = last_space
			assert split_at <= MAX_TEXT_LENGTH
			self.text = text[:split_at]
			next_text = text[split_at+1:] # omit the space or newline.
			logger.debug("Splitted {num} character long message text to {len_a} and {len_b}.".format(
				num=len(text), len_a=len(self.text), len_b=len(next_text)))
			self._next_msg = TextMessage(next_text, receiver)
"""

def is_sentence_end(char):
	return char in ".?!"

def is_sentence_part_end(char):
	return char in ",:;"

def is_word_separator(char):
	return char in [" ","\n",".",",","?","!","@","#","$",":",";","-","<",">","[","]","(",")","{","}","=","/","+","%","&","^","*","'","\"","`","~","|"]


def escape(string):
	for i in range(0, 7):
		string = string.replace(CHARS_UNESCAPED[i], CHARS_ESCAPED[i])
	return string
